"""
noid
====
Desired behaviours:

# generates a noid using arg defaults
noid

# generates a noid according to config
noid -c path/to/conf/file

# validate a noid
noid -v/--validate <noid>
# validating: <noid>
# valid/invalid

# overwrite scheme
noid -s/--scheme "doi:/"

# set naa
noid -N/--naa 1234

# overwrite template
noid -t/--template zeek

# set n
noid -n/--index

"""
import random
import re
import unittest

from noid import cli, pynoid, utils

import pathlib

BASE_DIR = pathlib.Path(__file__).parent.parent
CONFIG_FILE = BASE_DIR / 'noid' / 'noid.cfg'


class PynoidCLI(unittest.TestCase):
    def test_default(self):
        args = cli.cli(f"noid")
        self.assertIsNone(args.noid)
        self.assertIsNone(args.config_file)
        self.assertFalse(args.validate)
        self.assertEqual(args.scheme, cli.DEFAULT_SCHEME)
        self.assertEqual(args.naa, cli.DEFAULT_NAA)
        self.assertEqual(args.template, cli.DEFAULT_TEMPLATE)
        self.assertEqual(-1, args.index)
        self.assertFalse(args.verbose)

    def test_validate(self):
        """Validation requires noid positional argument"""
        self.assertIsNone(cli.cli(f"noid -v"))

    def test_config(self):
        """Using a config file"""
        args = cli.cli(f"noid -c {CONFIG_FILE}")
        self.assertIsNone(args.noid)
        self.assertEqual(str(CONFIG_FILE), args.config_file)
        self.assertFalse(args.validate)
        self.assertEqual('http://', args.scheme)
        self.assertEqual('83812', args.naa)
        self.assertEqual('zeeeeddddk', args.template)
        self.assertEqual(-1, args.index)
        self.assertFalse(args.verbose)


class PynoidAPI(unittest.TestCase):
    def test_mint(self):
        """Default use"""
        args = cli.cli(f"noid")
        noid = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        self.assertRegex(noid, r"^ark[:][/][\w\d]+")

    def test_mint_index(self):
        """Create from an index"""
        index = random.randint(0, 1000)
        args = cli.cli(f"noid -n {index}")
        noid1 = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        noid2 = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        self.assertEqual(noid1, noid2)

    def test_mint_scheme(self):
        """Set the scheme"""
        index = random.randint(1000, 9000)
        args = cli.cli(f"noid --scheme doi: --index {index}")
        noid = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        self.assertRegex(noid, r"^doi[:][\w\d]+")
        # the scheme should not affect the noid
        args2 = cli.cli(f"noid --index {index}")
        noid2 = pynoid.mint(template=args2.template, n=args2.index, scheme=args2.scheme, naa=args2.naa)
        # use regex to extract noids
        _noid = re.match(r"doi[:](?P<noid>[\w]+)", noid).group('noid')
        _noid2 = re.match(r"ark[:][/](?P<noid>[\w]+)", noid2).group('noid')
        self.assertEqual(_noid, _noid2)

    def test_mint_template_with_prefix(self):
        """Template can have one or more prefixes"""
        args = cli.cli(f"noid --template empiar.dddddk")
        noid = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        self.assertRegex(noid, r".*empiar\.[\w\d]+")

    def test_mint_with_naa(self):
        """Include naa"""
        naa = random.randint(9999, 99999)
        args = cli.cli(f"noid --naa {naa}")
        noid = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        self.assertRegex(noid, rf"^ark[:][/]{naa}[/][\w\d]+")

    def test_mint_combined(self):
        """All options at once"""
        naa = random.randint(9999, 99999)
        index = random.randint(9999, 99999)
        args = cli.cli(f"noid --scheme https:// --naa {naa} --index {index} --template zeeeeek")
        noid = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        self.assertRegex(noid, rf"^https[:][/]{{2}}{naa}[/][\w\d]+")

    def test_mint_invalid_template(self):
        """Invalid template should return blank noid"""
        args = cli.cli(f"noid --template abcdefg")
        noid = pynoid.mint(template=args.template, n=args.index, scheme=args.scheme, naa=args.naa)
        self.assertEqual('', noid)


class PynoidUtils(unittest.TestCase):
    def test_validate_mask(self):
        """Validate a mask"""
        self.assertTrue(utils.validate_mask('zek'))
        self.assertTrue(utils.validate_mask('ze'))
        self.assertTrue(utils.validate_mask('zdk'))
        self.assertTrue(utils.validate_mask('zd'))
        self.assertTrue(utils.validate_mask('zededededdeeddk'))
        self.assertTrue(utils.validate_mask('zededededdeedd'))
        self.assertTrue(utils.validate_mask('rek'))
        self.assertTrue(utils.validate_mask('re'))
        self.assertTrue(utils.validate_mask('rdk'))
        self.assertTrue(utils.validate_mask('rd'))
        self.assertTrue(utils.validate_mask('rededededdeeddk'))
        self.assertTrue(utils.validate_mask('rededededdeedd'))
        self.assertTrue(utils.validate_mask('sek'))
        self.assertTrue(utils.validate_mask('se'))
        self.assertTrue(utils.validate_mask('sdk'))
        self.assertTrue(utils.validate_mask('sd'))
        self.assertTrue(utils.validate_mask('sededededdeeddk'))
        self.assertTrue(utils.validate_mask('sededededdeedd'))
        self.assertTrue(utils.validate_mask('ek'))
        self.assertTrue(utils.validate_mask('e'))
        self.assertTrue(utils.validate_mask('dk'))
        self.assertTrue(utils.validate_mask('d'))
        self.assertTrue(utils.validate_mask('ededededdeeddk'))
        self.assertTrue(utils.validate_mask('ededededdeedd'))
        self.assertFalse(utils.validate_mask('a'))
        self.assertFalse(utils.validate_mask('aa'))
        self.assertFalse(utils.validate_mask('zeeedddl'))
        self.assertFalse(utils.validate_mask('zeeedtddl'))
        self.assertFalse(utils.validate_mask('zeeedtdd'))
        self.assertFalse(utils.validate_mask('adddeeew'))

    def test_get_noid_range(self):
        """Get the max_size"""
        xsize = len(utils.XDIGIT)
        dsize = len(utils.DIGIT)
        self.assertEqual(xsize ** 2 * dsize, utils.get_noid_range('zedek'))
        self.assertEqual(dsize ** 4, utils.get_noid_range('zddddk'))
        self.assertEqual(xsize ** 4, utils.get_noid_range('zeeeek'))
        self.assertEqual(xsize ** 3, utils.get_noid_range('seee'))
        self.assertEqual(dsize ** 2 * xsize ** 2, utils.get_noid_range('rddee'))

    def test_remove_prefix(self):
        """Remove the prefix"""
        self.assertEqual(('something.', 'eedeede'), utils.remove_prefix('something.eedeede'))
        self.assertEqual(('', 'eedeede'), utils.remove_prefix('eedeede'))
        self.assertEqual(('something.', 'eedeedek'), utils.remove_prefix('something.eedeedek'))
        self.assertEqual(('', 'eedeedek'), utils.remove_prefix('eedeedek'))
        self.assertEqual(('something.', 'reddek'), utils.remove_prefix('something.reddek'))
        self.assertEqual(('', 'reddek'), utils.remove_prefix('reddek'))


class PynoidTests(unittest.TestCase):

    def test_naa_append(self):
        noid = pynoid.mint(naa='abc')
        self.assertTrue(noid.startswith('abc/'))

    def test_scheme_append(self):
        schemes = ['doi:', 'ark:/', 'http://']
        for scheme in schemes:
            noid = pynoid.mint(scheme=scheme)
            self.assertTrue(noid.startswith(scheme))

    def test_mint_short_term(self):
        noid = pynoid.mint()
        self.assertTrue(noid.startswith(utils.SHORT))

    def test_mint_ns(self):
        """Over the range of each digit-space the index produces the respective digit"""
        ns = range(len(utils.DIGIT))
        for n in ns:
            self.assertEqual(utils.DIGIT[n], pynoid.mint('d', n))
        ns = range(len(utils.XDIGIT))
        for n in ns:
            self.assertEqual(pynoid.mint('e', n), utils.XDIGIT[n])

    def test_namespace_overflow(self):
        """Overflow occurs when we require a value outside of the range of the template"""
        self.assertEqual('', pynoid.mint(template='d', n=len(utils.DIGIT) + 1))
        self.assertEqual('', pynoid.mint(template='e', n=len(utils.XDIGIT) + 1))

    def test_mint_z_rollover(self):
        """Rollover happens when we exhaust the character space"""
        self.assertEqual(pynoid.mint('zd', len(utils.DIGIT)), '10')
        self.assertEqual(pynoid.mint('ze', len(utils.XDIGIT)), '10')

    def test_validate_valid(self):
        """Given some valid noids check that they are valid"""
        valid = 'test31wqw0ws8'
        valid_scheme = 'ark:/test31wqw0ws8'
        self.assertTrue(pynoid.validate(valid))
        self.assertTrue(pynoid.validate(valid_scheme))

    def test_validate_invalid(self):
        """Invalid noids with/out scheme"""
        invalid = 'test31qww0wsr'
        invalid_scheme = 'ark:/test31qww0wsr'
        self.assertFalse(pynoid.validate(invalid))
        self.assertFalse(pynoid.validate(invalid_scheme))

    def test_checkdigit(self):
        """The check digit is sensitive to permutations"""
        self.assertEqual('1Kw', pynoid.mint('eek', 100))
        self.assertFalse(pynoid.validate('K1w'))


if __name__ == '__main__':
    unittest.main()



import argparse
import shlex
import sys
from configparser import ConfigParser, ExtendedInterpolation

DEFAULT_NAA = ''
DEFAULT_SCHEME = 'ark:/'
DEFAULT_TEMPLATE = 'zeeddk'

# a global variable
parser = argparse.ArgumentParser(
    prog='noid',
    description='generate nice and opaque identifiers'
)
parser.add_argument(
    'noid',
    nargs='?',
    default=None,
    help="a noid"
)
parser.add_argument(
    '-c', '--config-file',
    help="path to a config file with a noid section"
)
parser.add_argument(
    '-v', '--validate',
    action='store_true',
    default=False,
    help="validate the given noid [default: False]"
)
parser.add_argument(
    '-s', '--scheme',
    default=DEFAULT_SCHEME,
    help=f"the noid scheme [default: '{DEFAULT_SCHEME}']"
)
parser.add_argument(
    '-N', '--naa',
    default=DEFAULT_NAA,
    help=f"the name assigning authority (NAA) number [default: {DEFAULT_NAA}]"
)
parser.add_argument(
    '-t', '--template',
    default=DEFAULT_TEMPLATE,
    help=f"the template by which to generate noids [default: '{DEFAULT_TEMPLATE}']"
)
parser.add_argument(
    '-n', '--index',
    type=int,
    default=-1,
    help="a number for which to generate a valid noid [default: random positive integer]"
)
parser.add_argument(
    '--verbose',
    action='store_true',
    default=False,
    help="turn on verbose text [default: False]"
)


class _ConfigParser(ConfigParser):
    """String-printable version"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        string = ""
        for section in self.sections():
            string += f"[{section}]\n"
            for option in self[section]:
                string += f"{option} = {self.get(section, option, raw=True)}\n"
            string += "\n"
        return string


def read_configs(args):
    configs = _ConfigParser(interpolation=ExtendedInterpolation())
    configs.read(args.config_file)
    return configs


def parse_args():
    """Parse CLI args"""
    args = parser.parse_args()
    # attach configs to the args namespace
    if args.config_file:
        if args.verbose:
            print(f"using configs: {args.config_file}", file=sys.stderr)
        configs = read_configs(args)
        _options = ['template', 'scheme', 'naa']
        if 'noid' in configs.sections():
            # overwrite using whatever is available
            for o in _options:
                if o in configs['noid']:
                    setattr(args, o, configs.get('noid', o))
                else:
                    print(f"warning: configs missing option '{o}'; using default value ({getattr(args, o)})", file=sys.stderr)
        else:
            print(f"warning: config file '{args.config_file}' lacks 'noid' section; ignoring config file",
                  file=sys.stderr)
    # argument validation
    if args.validate and args.noid is None:
        print("error: missing noid to validate", file=sys.stderr)
        return None
    return args


def cli(cmd):
    sys.argv = shlex.split(cmd)
    return parse_args()



DIGIT = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']  # 10
XDIGIT = DIGIT + ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'v', 'w',
                  'x', 'y', 'z'] + \
         ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'L', 'M', 'N', 'P', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
          'Z']
GENTYPES = ['r', 's', 'z']
DIGTYPES = ['d', 'e']
CHECKDIG = ['k']
SHORT = ''


#
# SHORT = '.shrt.'


def get_noid_range(mask):
    """Given the specified mask compute the maximum number of noids availabl

    :param str mask: the mask; if GENTYPE and CHECKDIG are present they will be ignored; only DIGTYPES are considered
    :return int max_int: the maximum number of noids
    """
    max_int = 1
    for c in mask:
        if c == 'e':
            max_int *= len(XDIGIT)
        elif c == 'd':
            max_int *= len(DIGIT)
    return max_int


def validate_mask(mask):
    """Check to make sure that we have a valid mask

    :param list mask: a sequence of characters
    :return bool: whether or not the mask is valid
    """
    # check the first character
    if not (mask[0] in GENTYPES or mask[0] in DIGTYPES):
        return False
    # check the last character
    elif not (mask[-1] in CHECKDIG or mask[-1] in DIGTYPES):
        return False
    # check all other characters
    else:
        for maskchar in mask[1:-1]:
            if not (maskchar in DIGTYPES):
                return False

    return True


def remove_prefix(template):
    """Split the template into the prefix and mask

    :param str template: a template string
    :return: the prefix
    :rtype str:
    :return: the mask
    :rtype str:
    """
    if '.' in template:
        prefix, mask = template.rsplit('.', 1)
        prefix += '.'
    else:
        mask = template
        prefix = ''
    return prefix, mask



import os
import sys
from random import randint

from noid import utils, cli


def mint(template='zek', n=-1, scheme='', naa='') -> str:
    """ Mint identifiers according to template with a prefix of scheme + naa.

    :param str template: a string consisting of GENTYPE + (DIGTYPE)+ [+ CHECKDIGIT]
    :param int n: a number to convert to a noid; default is -1 meaning create from random number
    :param str scheme: a scheme e.g. 'ark:/', 'doi:', 'http://', 'https://' etc.
    :param str naa: name assigning authority (number); can also be a string
    :return noid: a valid noid with/out check digit or the empty string (failure)
    :rtype str

    Template is of form [mask] or [prefix].[mask] where prefix is any URI-safe string and mask is a string of any
    combination 'e' and 'd', optionally beginning with a mint order indicator ('r'|'s'|'z') and/or ending
    with a checkdigit ('k'):

    Example Templates:
    d      : 0, 1, 2, 3
    zek    : 00, xt, 3f0, 338bh
    123.zek: 123.00, 123.xt, 123.3f0, 123.338bh
    seddee : 00000, k50gh, 637qg
    seddeek: 000000, k06178, b661qj

    The result is appended to the scheme and naa as follows: scheme + naa + '/' + [id].

    There is no checking to ensure ids are not reminted. Instead, minting can be controlled by supplying a (int)
    value for 'n'. It is possible to implement ordered or random minting from available ids by manipulating this
    number from another program. If no 'n' is given, minting is random from within the namespace. An indicator is
    added between '/' and [id] to mark these ids as for short term testing only. An override may be added later to
    accommodate applications which don't mind getting used ids.

    A note about 'r', 's', and 'z': 'z' indicates that a namespace should expand on its first element to accommodate
    any 'n' value (eg. 'de' becomes 'dde' then 'ddde' as numbers get larger). That expansion can be handled by this
    method. 'r' and 's' (typically meaning 'random' and 'sequential') are recognized as valid values, but ignored
    and must be implemented elsewhere.
    """
    # todo: handle 'r' and 's'
    prefix, mask = utils.remove_prefix(template)
    if not utils.validate_mask(mask):
        return ''
    if naa:
        naa += '/'
    _noid = generate_noid(mask, n)
    noid = f"{scheme}{naa}{prefix}{_noid}"
    if mask[-1] in utils.CHECKDIG:
        noid = f"{noid}{calculate_check_digit(_noid)}"
    return noid


def generate_noid(mask:str, n:int) -> str:
    """The actual noid generation

    :param str mask: the mask string
    :param int n: the number to use (default: -1, random number)
    :return str: the noid or an empty string
    """
    if n < 0:
        if mask[0] in utils.GENTYPES:
            mask = mask[1:]
        n = randint(0, utils.get_noid_range(mask) - 1)
    length = n
    noid = ''
    # construct the noid starting from the right
    # convert the numerical value 'n' to the digits specified in the mask
    for char in mask[::-1]:
        if char == 'e':
            div = len(utils.XDIGIT)
        elif char == 'd':
            div = len(utils.DIGIT)
        else:
            continue
        value = n % div
        n = n // div
        noid += (utils.XDIGIT[value])
    # if we have anytheng left over we continue using the leftmost mask character
    # this should not happen with 'z'
    if mask[0] == 'z':
        char = mask[1]
        while n > 0:
            if char == 'e':
                div = len(utils.XDIGIT)
            elif char == 'd':
                div = len(utils.DIGIT)
            else:
                print(f"error: template mask is corrupt; cannot process character: {char}", file=sys.stderr)
                return ''
            value = n % div
            n = n // div
            noid += (utils.XDIGIT[value])

    # if there is still something left over, we've exceeded our namespace.
    # checks elsewhere should prevent this case from ever evaluating true.
    if n > 0:
        print(f"error: cannot mint a noid for (counter = {length}) within this namespace.", file=sys.stderr)
        return ''
    # since we generated the noid from right to left we reverse it
    return noid[::-1]


def validate(noid: str) -> bool:
    """Checks if the final character is a valid checkdigit for the id. Will fail for ids with no checkdigit.

    This will also attempt to strip scheme strings (eg. 'doi:', 'ark:/') from the beginning of the string.
    This feature is limited, however, so if you haven't tested with your particular namespace,
    it's best to pass in ids with that data removed.

    :param str noid: a noid to validate
    :rtype bool: whether or not the noid is valid
    """
    return calculate_check_digit(noid[0:-1]) == noid[-1]


def calculate_check_digit(noid: str) -> str:
    """Given a noid determine the check digit to be appended from the alphabet

    :param str noid: a valid noid string
    :return str: a single character that is a check digit for the noid
    """
    # TODO: Fix checkdigit to autostrip scheme names shorter or longer than 3 chars.
    try:
        # if we have 'ark:/' remove them
        if noid[3] == ':':
            noid = noid[4:].lstrip('/')
    except IndexError:
        pass

    def index_of(x):
        """get the index of the extended digit"""
        try:
            return utils.XDIGIT.index(x)
        except:
            return 0

    total = list()
    for i, index in enumerate(map(index_of, noid)):
        total += [index * (i + 1)]
    index = sum(total) % len(utils.XDIGIT)
    return utils.XDIGIT[index]


def main():
    """Main entry point"""
    args = cli.parse_args()
    if args is None:
        return os.EX_USAGE
    if args.validate:
        if args.verbose:
            print(f"info: validating '{args.noid}'...", file=sys.stderr)
        print(f"'{args.noid}' valid? {validate(args.noid)}")
    else:
        if args.verbose:
            print(f"info: generating noid using template={args.template}, n={args.index}, "
                  f"scheme={args.scheme}, naa={args.naa}...", file=sys.stderr)
        noid = mint(args.template, args.index, scheme=args.scheme, naa=args.naa)
        print(noid)
    return os.EX_OK


if __name__ == '__main__':
    sys.exit(main())



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="noid",
    version="1.0.1",
    author="Paul K. Korir",
    author_email="pkorir@ebi.ac.uk, paulkorir@gmail.com",
    description="Mint NOIDs using a CLI or API",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/paulkorir/noid",
    packages=setuptools.find_packages(),
    package_data={
      'noid': ['noid.cfg']
    },
    entry_points={
        'console_scripts': [
            'noid=noid.pynoid:main',
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
